/*=======================================================================+
|             DEPARTAMENTO MECANICA UNIVERSIDADE AVEIRO                  |
|               Projecto Humanoide (PHUA - MODULROB)                     |
|                                                                        |
| Programadores:                                                         |
|              David Manuel Costa Gameiro          N:20281               |
|              Filipe Carvalho Mostardinha         N:22085               |
|                                                                        |
| Nome Ficheiro: MasterSec.c                                             |
|                                                                        |
|                            SOURCE FILE                                 |
+=======================================================================*/

#include "usart.h"
#include "can.h"

// Baud Rate das comunicaes pela USART
#define BAUD	115200

//////////////////////////////////////////
//   Mensagens tipo na recepcao         //
//////////////////////////////////////////

// _____ PACOTE DE TIPO DAS MENSAGENS VINDAS DO PC _______________________________________________________
// Mscaras para leitura dos diversos campos do pacote de tipo
enum MESSAGE_TYPE_PACKET  {
    // Campos do pacote de tipo
    OPERATION_MASK = 0b11000000,     // 0b11xxxxxx: 2 bits indicando o tipo de msg
    SCUID_MASK     = 0b00111100,     // 0bxx1111xx: 4 bits para a identificao do SCU
    PARAM_MASK     = 0b00000011,     // 0bxxxxxx11: 2 bits para a identificao do parmetro a ler/aplicar
    SENSORS_MASK   = 0b00000010,     // 0bxxxxxx1x: 1 bit para identificao do conjunto de sensores a ler
    FOOT_MASK      = 0b00000001,     // 0bxxxxxxx1: 1 bit para identificao do do p

    // Localizao dos vrios campos no pacote de tipo
    OPERATION_LOC  = 6,              // 0b11XXXXXX
    SCUID_LOC      = 2,              // 0bXX1111XX
    PARAM_LOC      = 0,              // 0bXXXXXX11
    SENSORS_LOC    = 1,              // 0bXXXXXX1X
    FOOT_LOC       = 0,              // 0bXXXXXXX1

    // Tipos de operaes (identificado pelos bits 0b11xxxxxx)
    OP_APPLY_JOINT  = 0b00,           // Aplicao de novas posies/velocidades/correntes a uma junta
    OP_READ_JOINT   = 0b01,           // Leitura das posies/velocidades/correntes de uma junta
    OP_READ_SENSORS = 0b10,           // Leitura dos sensores adicionais
    OP_READ_EXTBUFF = 0b11,           // Leitura do buffer externo

    // Tipos de parmetros (identificado pelos bits 0bxxxxxx11)
    PARAM_POSITION  = 0b00,           // Posio
    PARAM_VELOCITY  = 0b01,           // Velocidade
    PARAM_CURRENT   = 0b10,           // Corrente
    PARAM_CONTROL   = 0b11,           // Parmetros do controlador

    // Conjuntos de sensores (identificado pelo bit 0bxxxxxx1x
    SENSORS_REACTION    = 0,
    SENSORS_ORIENTATION = 1
};

// Macros para leitura dos diversos campos no pacote de tipo
#define operation(x) (((unsigned int) (x) & OPERATION_MASK) >> OPERATION_LOC)   // Tipo de mensagem
#define scuID(x)     (((unsigned int) (x) & SCUID_MASK    ) >> SCUID_LOC    )   // Identificao do SCU
#define param(x)     (((unsigned int) (x) & PARAM_MASK    ) >> PARAM_LOC    )   // Identificao do parametro
#define sensors(x)   (((unsigned int) (x) & SENSORS_MASK  ) >> SENSORS_LOC  )   // Identificao do conjunto de sensores
#define footID(x)    (((unsigned int) (x) & FOOT_MASK     ) >> FOOT_LOC     )   // Identificao do p

// Macros para construo dos diversos padres no pacote de tipo
#define operation_patt(y)   (byte) (((y) << OPERATION_LOC) & OPERATION_MASK)   // Tipo de mensagem
#define scuID_patt(y)       (byte) (((y) << SCUID_LOC    ) & SCUID_MASK    )   // Identificao do SCU
#define param_patt(y)       (byte) (((y) << PARAM_LOC    ) & PARAM_MASK    )   // Identificao do parametro
#define sensors_patt(y)     (byte) (((y) << SENSORS_LOC  ) & SENSORS_MASK  )   // Identificao do conjunto de sensores
#define footID_patt(y)      (byte) (((y) << FOOT_LOC     ) & FOOT_MASK     )   // Identificao do p

// Macro para construo do pacote de tipo da mensagem
#define typePacket(op,more)             (operation_patt(op) | more)
#define typePacketJoint(op,scu_id,wh)   (operation_patt(op) | scuID_patt(scu_id) | param_patt(wh))
#define typePacketSensors(sens,foot)    (operation_patt(OP_READ_SENSORS) | sensors_patt(sens) | footID_patt(foot))
// ___________________________________________________________________________________________________________

// _____ PACOTE DE ESTADO DO SISTEMA _________________________________________________________________________
// Pacote de estado do sistema (ltimo byte das mensagens OP_READ_JOINT
enum MESSAGE_SYSSTATUS_PACKET {
    // Mscaras dos bits
    DEADLINEERROR_MASK = 0b00000100,     // Ocorrncia de violao de deadline
    CONTROLON_MASK     = 0b00000010,     // Controlador PID ligado
    RESET_MASK         = 0b00000001,     // Motores nas home positions

    // Localizao dos bits
    DEADLINEERROR_LOC  = 2,
    CONTROLON_LOC      = 1,
    RESET_LOC          = 0
};

// Construes dos padres do pacote de estado do sistema
#define deadlineError_patt(d) (byte) (((unsigned int) (d) << DEADLINEERROR_LOC) & DEADLINEERROR_MASK)
#define controlOn_patt(c)     (byte) (((unsigned int) (c) << CONTROLON_LOC    ) & CONTROLON_MASK    )
#define reset_patt(r)         (byte) (((unsigned int) (r) << RESET_LOC        ) & RESET_MASK        )

// Construo do pacote de estado do sistema
#define sysStatusPacket(res,cOn,d)   (reset_patt(res) | controlOn_patt(cOn) | deadlineError_patt(d))
// _________________________________________________________________________________________________________

// Buffer de software da USART
static struct  {
    byte rx[USART_RECV_LENGTH]; // Buffer de recepo
    byte tx[USART_SEND_LENGTH]; // Buffer de transmisso
    byte tx_i, rx_i;            // ndices de acesso aos buffers
    bool txSemaph;              // Semforo de bloqueio  modificao do buffer de transmisso
} usartBuffer;

// Buffer de segundo nvel, apelidado por externo
static struct  {
    byte tx[USART_EXTBUFF_LENGTH];
    bool req;
} extBuffer;

// Valor inicial do Timer0 para tentativa de retransmisso de mensagens
const unsigned char kt0 = (unsigned char) (256-(unsigned int) T0_OUT*F_CPU/T0_PRES);

//*****************************************************************************
static void initUsartBuffers(void);                     // Inicializao dos buffers da USART

// Inicializacao da USART
//
void usartInit()
{
	float spbrg;
	
    // Inicializao dos buffers (de software) da USART
    initUsartBuffers();

    // Programacao dos pinos RC7 e RC6 como entrada e sada respectivamente
	TRISC=(TRISC | 0x80) & 0xBF;
	
    // Clculo do valor associado ao baudrate pedido
	spbrg=((float) F_CPU*1e6)/((float) 4*BAUD)-1.0;
	if (spbrg-(unsigned char) spbrg>=0.5)  spbrg+=1.0;

    // Programacao da Usart (Assincrona por Polling)
	SPBRG=(unsigned char) spbrg;    // BaudRate
	TXSTAbits.SYNC = 0;             // Modo assincrono
    TXSTAbits.BRGH = 1;	            // BR high speed
	TXSTAbits.TXEN = 1;	            // Transmit Enable 
    RCSTAbits.CREN = 1;	            // Continuous Reception Enable
	RCSTAbits.SPEN = 1;	            // Serial Port Enable

    PIE1bits.RCIE=1;                // (Des)Activao das interrupes
    PIE1bits.TXIE=0;

    // Configurao do Timer0 para gerao de uma interrupo aps 100ms
    // Este interrupt ser usado para a tentativa de retransmisso de uma
    // mensagem agendada.

	T0CONbits.T08BIT=1;		// 8 bit timer
    T0CONbits.T0CS=0;       // Internal instruction cycle clock

    T0CONbits.PSA=0;        // Prescaler is assigned
    T0CONbits.T0PS2=0;      // Prescaler => 1:8
    T0CONbits.T0PS1=1;
    T0CONbits.T0PS0=0;

	T0CONbits.TMR0ON=1;		// Timer0 is On

    INTCONbits.TMR0IE=0;    // Interrupt Disable

    // Activao das interrupes globais, no mduloPIC.H/C
}

//*****************************************************************************
// Inicializao dos buffers da Usart
//
static void initUsartBuffers(void)
{
    byte i;

    // Inicializao a zero de toda a estrutura usartBuffer
    for (i=0; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
    for (i=0; i<USART_RECV_LENGTH; ++i)  usartBuffer.rx[i]=0;

    usartBuffer.tx_i=usartBuffer.rx_i=0;

    // Transmisso por interrupo autorizada
    usartBuffer.txSemaph=true;

    // Inicializao do buffer externo
    for (i=0; i<USART_EXTBUFF_LENGTH; ++i)  extBuffer.tx[i]=0;
    extBuffer.req=false;
}

// _____ FUNES DE MANIPULAO DIRECTA DO BUFFER RX/TX __________________
//*****************************************************************************
// Retorna true, se o buffer de recepo est a receber um stream de bytes
//
bool usartStreamMode(void)
{
    return (usartBuffer.rx_i>0);
}

//*****************************************************************************
// Insero de um byte no buffer de recepo.
// Retorna true, se a insero teve sucesso
// *full indica se o buffer de recepo est cheio
//
bool usartStoreRx(byte val, bool *full)
{
    // O buffer rx est cheio?
    if (usartBuffer.rx_i>=USART_RECV_LENGTH)  {
        if (full!=(bool *) 0)  *full=true;      // Buffer cheio!
        return false;                           // Insero falhada
    }

    // Insero do novo byte
    usartBuffer.rx[usartBuffer.rx_i++]=val;

    // Verificao de buffer cheio
    if (full!=(bool *) 0)  {
        if (usartBuffer.rx_i==USART_RECV_LENGTH)  *full=true;
        else                                      *full=false;
    }
    // Insero foi um sucesso
    return true;
}

//*****************************************************************************
// Obteno e remoo de um byte do buffer de transmisso.
// *var conter o byte removido;
// *empty indica se o buffer de transmisso est vazio
// True  retornado se a obteno teve sucesso.
//
bool usartGetTx(byte *var, bool *empty)
{
    // Verificando se o buffer de transmisso est vazio
    if (usartBuffer.tx_i>=USART_SEND_LENGTH)  {
        if (empty!=(bool *) 0)  *empty=true;    // Buffer de transmisso vazio
        return false;                           // Obteno do byte no efectuada
    }

    // Obteno do byte
    if (var!=(byte *) 0)  *var=usartBuffer.tx[usartBuffer.tx_i++];

    // Verificando se o buffer de transmisso ficou vazio
    if (empty!=(bool *) 0)  {
        if (usartBuffer.rx_i==USART_RECV_LENGTH)  *empty=true;
        else                                      *empty=false;
    }
    // Obteno efectuada com sucesso
    return true;
}

//*****************************************************************************
// Reset do Buffer de recepo.
// O buffer  esvaziado e pronto para a recepo de novos caracteres.
//
void usartResetRxBuff(void)
{
    usartBuffer.rx_i=0;
}

//*****************************************************************************
// Reset do Buffer de transmisso.
// O buffer  esvaziado e pronto para armazenar novos bytes.
//
void usartResetTxBuff(void)
{
    usartBuffer.tx_i=0;
}

//*****************************************************************************
// Leitura do estado do semforo de bloqueio no acesso ao
// buffer de transmisso/externo.
//
bool usartBufftxSemaph(void)
{
    return usartBuffer.txSemaph;
}

// _____ FUNES PARA COMUNICAO LIVRE _______________________________________
//*****************************************************************************
// Rotina para o envio imediato de uma mensagem pela USART.
// Este modo de envio NO segue a poltica Request->Send; Simplesmente envia a mensagem!
//
void usartSendMsgNOW(byte *message, byte n)
{
    byte i;

    // Bloqueio no acesso ao buffer de transmisso
    usartBuffer.txSemaph=false;

    // Esperando que a USART fique livre para a transmisso
    while (PIE1bits.TXIE);

    // Neste ponto, j podemos enviar uma mensagem.

    // Caso a mensagem a enviar seja demasiado grande, trunc-la.
    if (n>USART_SEND_LENGTH)  n=USART_SEND_LENGTH;

    // Preenchimento do buffer de transmisso com os dados a enviar
    for (i=0; i<n; ++i)  usartBuffer.tx[i]=message[i];

    // Preenchimento do resto da mensagem com zeros
    for (i=n; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;

    // Incio de transmisso
    PIE1bits.TXIE=1;

    // Desbloqueio no acesso ao buffer de transmisso
    usartBuffer.txSemaph=true;
}

//*****************************************************************************
// Armazenamento de uma mensagem no buffer externo da USART.
// Este buffer  necessrio para a transmisso uma mensagem
// configurvel pelo utilizador, seguindo a poltica Request->Send!
//
void usartStore2extBuff(byte *message, byte n)
{
    byte i;

    // Bloqueio no acesso ao buffer de externo por parte de interrupes
    usartBuffer.txSemaph=false;

    // Caso a mensagem apresentada seja demasiado grande, trunc-la.
    if (n>USART_EXTBUFF_LENGTH)  n=USART_EXTBUFF_LENGTH;

    // Preenchimento do buffer externo
    for (i=0; i<n; ++i)  extBuffer.tx[i]=message[i];

    // Finalizao do preenchimento com zeros
    while (i<USART_EXTBUFF_LENGTH)  extBuffer.tx[i++]=0;

    // Desbloqueio no acesso ao buffer externo
    usartBuffer.txSemaph=true;
}

// _____ FUNES PARA USO DA RSI  _____________________________________________
// Importante: As funes abaixo modificam o contedo do buffer de
// transmisso sem considerar o semforo ou se j existe alguma
// transmisso em curso, pelo que no devem ser usadas de forma livre.
// Apenas se destinam  Rotina de Servio  interrupo!

//*****************************************************************************
// Envio de uma mensagem de teste (FA F9 F8 F7 F6 F5)
//
void usartSendTestMsg(void)
{
    byte i;

    for (i=0; i<USART_SEND_LENGTH; ++i)
        usartBuffer.tx[i]=MESSAGE_TEST-i;
}

//*****************************************************************************
// Envio de uma mensagem de status
//
void usartSendStatus(byte stat)
{
    byte i;

    // O pacote SOF conter o status a enviar:
    usartBuffer.tx[SOF]=stat;

    // Preenchimento do resto da mensagem com o contedo da ltima recebida
    for (i=TYPE; i<USART_RECV_LENGTH; ++i)  usartBuffer.tx[i]=usartBuffer.rx[i];

    // Preenchimento com zeros do resto da mensagem
    for (i=USART_RECV_LENGTH; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
}

//*****************************************************************************
static void newJointInfo(byte scu_id, byte param);      // Actualizao dos dados de uma junta
static void sendJointInfo(byte scu_id, byte param);     // Envio das informaes sensoriais de uma junta
static void sendReactionInfo(byte foot);                // Envio da matriz de sensores de fora de um p
static void sendExtraInfo(void);                        // Envio da informao dos sensores adicionais (inclinmetro e giroscpio)
static void sendExtBuffer(void);                        // Envio de uma mensagem com o contedo do buffer externo

// Processamento de uma mensagem recebida pela Usart (a partir do PC)
//
void usartProcessMsg(void)
{
    byte msg_type;

    // Padro descritor do tipo da mensagem
    msg_type=usartBuffer.rx[TYPE];

    // De acordo com a operao referente  mensagem...
    switch (operation(msg_type))  {

        // Aplicao de uma nova posio, velocidade ou corrente a todos os servos de uma junta
        case OP_APPLY_JOINT:    newJointInfo(scuID(msg_type),param(msg_type));
                                canSendMsg(HOST_MCU+1);
                                break;

        // Leitura de um parmetro de todos os servos de uma junta
        case OP_READ_JOINT:     sendJointInfo(scuID(msg_type),param(msg_type));
                                break;

        // Leitura dos sensores adicionais (sensores de fora, inclinmetro ou giroscpio)
        case OP_READ_SENSORS:   if (sensors(msg_type)==SENSORS_REACTION)
                                    sendReactionInfo(footID(msg_type));
                                else
                                    sendExtraInfo();
                                break;

        // Leitura do contedo do buffer externo
        case OP_READ_EXTBUFF:   sendExtBuffer();
                                break;
    }
}


//*****************************************************************************
// Aplicao de uma nova posio, velocidade ou corrente a todos os servos de uma junta
//
static void newJointInfo(byte scu_id, byte param)
{
    byte scu, servo;

    // Validao do ndice do SCU
    if ((scu=scuIndex(scu_id))>=N_SCU)  {
        usartSendStatus(MESSAGE_INVREQ);
        return;
    }

    // Preenchimento dos arrays internos com os novos dados recebidos (posio, velocidade ou corrente)
    for (servo=0; servo<N_SERVOS; ++servo)
        switch (param)  {
            // Posio
            case PARAM_POSITION:    actuators.servo[scu][servo].position=usartBuffer.rx[DATA1+servo];
                                    break;
            // Velocidade
            case PARAM_VELOCITY:    actuators.servo[scu][servo].velocity=usartBuffer.rx[DATA1+servo];
                                    break;
            // Corrente
            case PARAM_CURRENT:     actuators.servo[scu][servo].current =usartBuffer.rx[DATA1+servo];
                                    break;
            // Parmetros do controlador
            case PARAM_CONTROL:     actuators.control.kp=usartBuffer.rx[DATA1+0];
                                    actuators.control.kd=usartBuffer.rx[DATA1+1];
                                    break;
        }

    // Envio de uma mensagem indicadora que a operao pedida teve sucesso!
    usartSendStatus(MESSAGE_SUCESS);
}


//*****************************************************************************
// Leitura de um parmetro de todos os servos de uma junta
//
static void sendJointInfo(byte scu_id, byte param)
{
    byte i, servo, scu;

    // Validao do ndice do SCU
    if ((scu=scuIndex(scu_id))>=N_SCU)  {
        usartSendStatus(MESSAGE_INVREQ);
        return;
    }

    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;                                // Start of Frame
    usartBuffer.tx[TYPE]=typePacketJoint(OP_READ_JOINT,scu_id,param);   // Tipo da mensagem

    // Preenchimento do buffer de transmisso com as informaes sensoriais da junta
    for(servo=0; servo<N_SERVOS; ++servo)
        switch (param)  {
            // Posio
            case PARAM_POSITION:    usartBuffer.tx[DATA1+servo]=sensors.servo[scu][servo].position;
                                    break;
            // Velocidade
            case PARAM_VELOCITY:    usartBuffer.tx[DATA1+servo]=sensors.servo[scu][servo].velocity;
                                    break;
            // Corrente
            case PARAM_CURRENT:     usartBuffer.tx[DATA1+servo]=sensors.servo[scu][servo].current;
                                    break;
            // Outros
            default:                usartBuffer.tx[DATA1+servo]=0;
        }

    // Pacote de estado do sistema
    usartBuffer.tx[DATA1+N_SERVOS]=sysStatusPacket(sensors.sysStatus.reset,sensors.sysStatus.controlOn,sensors.sysStatus.deadlineError);

    // Preenchimento com zeros do resto da mensagem
    for (i=DATA1+N_SERVOS+1; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
}


//*****************************************************************************
// Leitura da matriz dos sensores de fora de um determinado p
//
static void sendReactionInfo(byte foot)
{
    byte i;

    // Validao do p pedido
    if (foot>=N_FEET)  {
        usartSendStatus(MESSAGE_INVREQ);
        return;
    }

    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;                            // Start Of Frame
    usartBuffer.tx[TYPE]=typePacketSensors(SENSORS_REACTION,foot);  // Tipo da mensagem

    // Preenchimento do buffer de transmisso com os valores dos sensores de fora
    for(i=0; i<N_FORCE_SENSORS; ++i)  usartBuffer.tx[DATA1+i]=sensors.reaction[foot][i];

    // Preenchimento com zeros do resto da mensagem
    for (i=DATA1+N_FORCE_SENSORS; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
}


//*****************************************************************************
static void sendExtraInfo(void)
{
    unsigned int i;

    // Preenchimento do buffer de transmisso
    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;                            // Start Of Frame
    usartBuffer.tx[TYPE]=typePacketSensors(SENSORS_ORIENTATION,0);  // Tipo da operao

    usartBuffer.tx[DATA1+0]=sensors.inclinometer[0];                // Dados do inclinmetro
    usartBuffer.tx[DATA1+1]=sensors.inclinometer[1];
    usartBuffer.tx[DATA1+2]=sensors.giroscope[0];                   // Dados do giroscpio
    usartBuffer.tx[DATA1+3]=sensors.giroscope[1];

    // Preenchimento do resto da mensagem com zeros
    for (i=DATA1+4; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
}


//*****************************************************************************
// Envio do contedo do buffer externo pela USART
//
static void sendExtBuffer(void)
{
    byte i;

    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;
    usartBuffer.tx[TYPE]=typePacket(OP_READ_EXTBUFF,0);

    // Cpia dos dados do buffer externo para o interno
    for (i=0; i<USART_EXTBUFF_LENGTH; ++i)  {
        usartBuffer.tx[DATA1+i]=extBuffer.tx[i];    // Transferncia da informao
        extBuffer.tx[i]=0;                          // Clear do buffer origem
    }

    // A flag req indica que os dados foram acessados pela USART
    extBuffer.req=true;
}


