//*******************************************************************************************
// SERVO.C
//
// Source file com a definio dos device drivers para o controlo de um servomotor.
//
// (c) Milton Ruas <a21824@alunos.det.ua.pt>
//     DETUA & DEMUA, Nov/2005
//*******************************************************************************************

#include "pic2.h"
#include "can.h"

// ****** CONSTANTES ************************************************************************

// O Timer 0  usado para fazer aquisio do sinal proveniente do potencimetro
#define T0_PRESC	4			// Prescaler (alterar o programa)
#define T0_OUT		100		    // Perodo do Timer (us)
static const unsigned char kt0 = (unsigned char) (256-(unsigned int) F_CPU*T0_OUT/T0_PRESC);

// O Timer 1  utilizado para definir o perodo do PWM para o servo
#define T1_PRES 4				// Prescaler seleccionado na programacao do timer 1 (alterar funo initPit(...))
#define T1_OUT	20				// Perodo do Timer 1 (ms)
#define KT1		(65535-((unsigned int) F_CPU*1000/T1_PRES)*T1_OUT+1) // Constante de diviso p/ o Timer 1
#define KT1_H	(unsigned char)(KT1 >> 8)
#define KT1_L	(unsigned char)(KT1 & 0xFF)

#define T2_PRES		4       // Prescaler do Timer 2 (alterar funo initPit(..))
#define T2_POST     9       // Postscaler do Timer 2 (alterar funo initPit(..))
#define T2_RES_FACTOR   10
#define T2_OUT_HF   (unsigned char) (((unsigned int) DUTY_MAX-DUTY_MIN)*T2_RES_FACTOR/(unsigned int) SERVO_RES)
#define KT2_1ST     (unsigned char) ((unsigned int) DUTY_MIN*F_CPU/(T2_PRES*T2_POST)-1)
#define KT2_HF      (unsigned char) ((unsigned int) T2_OUT_HF*F_CPU/T2_RES_FACTOR-1)
//static const byte KT2_1ST  =   (unsigned char) ((unsigned int) DUTY_MIN*F_CPU/(T2_PRES*T2_POST)-1);
//static const byte KT2_HF   =   (unsigned char) ((unsigned int) T2_OUT_HF*F_CPU/T2_RES_FACTOR-1);

// ****** VARIVEIS ESTTICAS E GLOBAIS *****************************************************

static struct  {
    unsigned char which;
    struct {
	    unsigned char vpotImp;
	    unsigned int  vpotMin, vpotMin_prev;
    } measure[N_SERVOS];
} servo;

// Variveis da estrutura de informao temporal
bool timeTick=false;

// ****** DEFINIO DAS FUNES *************************************************************
void highISR(void);						// Prottipo da funo de servio a interrupes
void lowISR(void);

#pragma code InterruptVectorHigh = 0x08	//Quando ocorre um interrupt o programa "salta" para este endereo
void InterruptVectorHigh (void)
{
    _asm
	goto highISR
    _endasm
}

#pragma code InterruptVectorLow = 0x18	//Quando ocorre um interrupt o programa "salta" para este endereo
void InterruptVectorLow (void)
{
    _asm
	goto lowISR
    _endasm
}

#pragma code

static void resetTimer2(void);
static void initVars(void);

// ==========================================================================================
// Device Drivers
// ==========================================================================================

// Inicializacoes relativas ao microcontrolador
// pos0  a posio inicial do servo (entre -90 e +90 graus)
//
void initPic(void)
{
    // Inicializao das variveis internas do servo
    initVars();

    // Programao dos portos de entrada/sada
	TRISA=0b00000001;		// RA0 input / RA1..RA7 outputs
	TRISB=0b00000000;		// RB0...RB7 outputs
	TRISC=0b00000000;		// RC0...RC7 outputs

    // Inicializao dos portos
    PORTA=0;                // Seleco do sinal a amostrar
    SERVO1_PIN=0;           // Pinos de PWM
    SERVO2_PIN=0;
    SERVO3_PIN=0;
    TESTPIN=0;

	// Programao do Timer 2 (desligado e interrupes desactivadas)

    resetTimer2();
   	T2CONbits.TMR2ON=1;		// Timer2 ligado
	IPR1bits.TMR2IP=1;		// High priority interrupt
   	PIE1bits.TMR2IE=0;      // Interrupes do Timer2 desactivadas

    // Programacao do Timer 1

	T1CONbits.RD16=0;		// read/write in two 8-bit operations

	T1CONbits.T1CKPS1 = 1;
	T1CONbits.T1CKPS0 = 0;	// prescaler 1:4

    T1CONbits.T1OSCEN=0;    // Oscillator is shut-off
    T1CONbits.TMR1CS=0;     // Internal clock (Fosc/4)

    TMR1H = KT1_H;  		// Period
	TMR1L = KT1_L;

	T1CONbits.TMR1ON=1;		// Timer 1 ON
	IPR1bits.TMR1IP=1;		// Low priority interrupt
	PIE1bits.TMR1IE=1;		// Timer1 interrupt enable

    // Programacao do Timer 0

	T0CONbits.T08BIT=1;		// 8 bit timer
    T0CONbits.T0CS=0;       // Internal instruction cycle clock

    T0CONbits.PSA=0;        // Prescaler is assigned
    T0CONbits.T0PS2=0;      // Prescaler => 1:4
    T0CONbits.T0PS1=0;
    T0CONbits.T0PS0=1;

    TMR0L=kt0;              // Contador inicializado
	T0CONbits.TMR0ON=1;		// Timer0 is On
	INTCON2bits.TMR0IP=1;	// High priority interrupt
    INTCONbits.TMR0IE=1;    // Interrupes activadas

    // Programacao da ADC

	ADCON0=0b10000001;      // Fosc/64, channel 0(RA0), enable ADC converter
	ADCON1=0b01001110;      // Left justified; RA0 analog; the rest are digital

	IPR1bits.ADIP=1;		// Low Priority interrupt
	PIE1bits.ADIE=0;		// Low Priority interrupt

	// Inicializao das comunicaes
	initCan();

    // Activacao global das interrupcoes
	RCONbits.IPEN=1;		// Interrupt priority enable
	INTCONbits.GIEH=1;		// Enable high priority interrupts
	INTCONbits.GIEL=1;      // Enable low priority interrupts
}

static void resetTimer2(void)
{
	T2CONbits.T2CKPS1 = 0;	// Prescaler de 1:4
	T2CONbits.T2CKPS0 = 1;

	T2CONbits.TOUTPS3 = 1;	// Postscaler de 1:9
	T2CONbits.TOUTPS2 = 0;
	T2CONbits.TOUTPS1 = 0;
	T2CONbits.TOUTPS0 = 0;

   	PR2  = KT2_1ST;			// Perodo standard
}

static void initVars(void)
{
    unsigned char srv;

    servo.which=0;

    for (srv=0; srv<N_SERVOS; ++srv)  {

        servo.measure[srv].vpotMin=0xFFFF;
        servo.measure[srv].vpotMin_prev=0;
        servo.measure[srv].vpotImp=0;
    }

    // Inicializao das estruturas externas
    initServoData();
}

// ==========================================================================================
// Funes temporais
// ==========================================================================================

// Funcao para gerar um atraso (ms)
//
void wait(unsigned int value)
{
	unsigned int i, end=value/T1_OUT;

	for (i=0; i<end; ++i)  {
		while (timeTick==0);
   		timeTick=0;
	}
}

// ==========================================================================================
// Funes relativas a interrupes
// ==========================================================================================

// Actualizao das medidas
static void updateServoMeasures(void);
static void finalizeServoMeasures(void);

// Rotina de servio  interrupo
//
#pragma interrupt highISR
void highISR(void)
{
	unsigned char counter;

    // Interrupo responsvel por baixar o PWM
    if (PIE1bits.TMR2IE && PIR1bits.TMR2IF)  {

        // O Timer2  redefinido de modo a provocar interrupes de alta frequncia
        T2CON=0b00000100;
        PR2=KT2_HF;

        counter=0;
        do  {
            while (PIR1bits.TMR2IF==0);
            PIR1bits.TMR2IF=0;
            
            if (SERVO1_PIN && counter>=pwmServo[0].pwm)  SERVO1_PIN=0;    // Desactivar PWM?
            if (SERVO2_PIN && counter>=pwmServo[1].pwm)  SERVO2_PIN=0;
            if (SERVO3_PIN && counter>=pwmServo[2].pwm)  SERVO3_PIN=0;

        } while (counter++<SERVO_RES);

        resetTimer2();
        PIE1bits.TMR2IE=0;      // Interrupes do Timer2 desactivadas

        counter=0;
    
        // Envio de mensagens com a informao sensorial
        sendCanMessage();

        // Actualizao dos time Ticks temporais
    	timeTick=true;

        // Reinicio das medies dos servos
        
        INTCONbits.TMR0IF=1;
        INTCONbits.TMR0IE=1;
	}

    if (INTCONbits.TMR0IE && INTCONbits.TMR0IF)  {
        // Arranque do timer0 para provocar um delay antes da prxima medio
        TMR0L=kt0;

        TESTPIN=1;

        // Arranque da ADC para leitura do sensor do servo
        PIR1bits.ADIF=0;
        PIE1bits.ADIE=1;
        ADCON0bits.GO=1;

        INTCONbits.TMR0IF=0;
    }

    if (PIE1bits.ADIE && PIR1bits.ADIF)  {

        PIR1bits.ADIF=0;
        PIE1bits.ADIE=0;

        // Actualizao das medidas
        updateServoMeasures();

        TESTPIN=0;
    }

    // Interrupo responsvel por elevar o PWM
    if (PIE1bits.TMR1IE && PIR1bits.TMR1IF)  {

        TMR1H=KT1_H;            // Inicializao do contador do Timer 1
        TMR1L=KT1_L;

        // Inicializao do Timer2 para gerar o primeiro interrupt de alta frequncia
        TMR2=0;
       	PIR1bits.TMR2IF=0;
       	PIE1bits.TMR2IE=1;		// Activao das interrupes do Timer2

        SERVO1_PIN=1;           // Output for Servos activated
        SERVO2_PIN=1;
        SERVO3_PIN=1;

        TESTPIN2=1;

        // Paragem na leitura dos sensores
        PIE1bits.ADIE=0;
        INTCONbits.TMR0IE=0;
        // Clculo dos valores sensoriais definitivos para um ciclo de PWM
        finalizeServoMeasures();
        // Transferncia dos resultados para as estruturas globais
        transferServoSensors();

        receiveCanMessage(0);
        receiveCanMessage(1);

        // Actualizao dos actuadores
        controller();

		PIR1bits.TMR1IF=0;      // Reset Timer1 interrupt flag
        TESTPIN2=0;
	}
}

#pragma interrupt lowISR
void lowISR(void)
{/*
    if (PIR3bits.RXB0IF || PIR3bits.RXB1IF)
        // Recepo de mensagem com os dados dos actuadores
        receiveCanMessage();*/
}

static void updateServoMeasures(void)
{
    // Leitura da tenso presente no potencimetro do servo
    unsigned int vpot=((unsigned int) ADRESH<<2) | (ADRESL>>6);

	if (vpot<servo.measure[servo.which].vpotMin)
        servo.measure[servo.which].vpotMin=vpot;  // Actualizao do valor mnimo

    // Medio da largura do impulso de corrente
	if (vpot>=servo.measure[servo.which].vpotMin_prev+VPOT_IMPULSE_LIMIAR)
        servo.measure[servo.which].vpotImp++;

    // Transferncia da entrada do multiplexer para o servo seguinte
    if (++servo.which>=N_SERVOS)  servo.which=0;
//    PORTA = servo.which << 1;
    PORTAbits.RA1= servo.which & 1;
    PORTAbits.RA2= (servo.which >> 1) & 1;
}


// Clculo dos valores sensoriais definitivos para um ciclo de PWM
//
static void finalizeServoMeasures(void)
{
    byte srv;
    word tmp;

  	// Passagem dos valores finais
    for (srv=0; srv<N_SERVOS; ++srv)  {

        // Truncagem da tenso para 2.50V (200 valores possveis)
        if ((tmp=servo.measure[srv].vpotMin)>511)  tmp=511;
        // Diminuio da resoluo para metade (100 valores possveis)
        tmp /= 2;
	    adcServo[srv].vpotMin = (byte) (tmp*250/255);

        // Pre-processamento do impulso de corrente
	    adcServo[srv].vpotImp=servo.measure[srv].vpotImp;
	
	    // Reinicializao das variveis auxiliares
	    servo.measure[srv].vpotMin_prev=servo.measure[srv].vpotMin;
	    servo.measure[srv].vpotMin=0xFFFF;
	    servo.measure[srv].vpotImp=0;
    }
}
