#include "servo.h"

// Macro conversora de ngulo para unidades de PWM
#define PWM(ang)    (byte) ((SERVO_RANGE-(signed int) (ang))*SERVO_RES/(2*(signed int) SERVO_RANGE))

// Estruturas externas
struct_pwmServo pwmServo[N_SERVOS];     // Dados dos actuadores
struct_velocity velocityServo[N_SERVOS]; // Dados dos sensores


// Inicializao das estruturas de dados
//
void initServoData(void)
{
    byte srv, i;

    for (srv=0; srv<N_SERVOS; ++srv)  {
        // Inicializao das posies dos servos nos valores standard
        pwmServo[srv].applyPosition=homePosition[srv];
        pwmServo[srv].pwm=PWM(homePosition[srv]);
        pwmServo[srv].control.angle=(signed int) homePosition[srv]*ANGLE_FACTOR;
        pwmServo[srv].control.error_prev=0;

        // Inicializao dos arrays de posio
        for (i=0; i<N_POS; ++i)  {
            velocityServo[srv].positions[i]=homePosition[srv];
            velocityServo[srv].oldPosition=0;
        }
    }
}

// Aplicao de uma nova velocidade a um servo
//
void servoPos(signed char position, byte srv)
{
    if (position<-SERVO_RANGE)       position=-SERVO_RANGE;
    else if (position>SERVO_RANGE)   position= SERVO_RANGE;

    pwmServo[srv].applyPosition=position;
}

// Funcao para aquisicao do sinal presente no potencimetro do servo
//
// Os seguintes valores so utilizados para converter uma tenso entre 0 e 5000 num ngulo em graus.
// POS(graus)=ORIGIN-<tenso>*SLOPE/QUOCIENT
#define SLOPE      9        // Declive=1800     => Simplificando 1800/10000=9/50: SLOPE=9
#define QUOCIENT   50       // Quociente=10000  => Simplificando 1800/10000=9/50: QUOCIENT=50
#define ORIGIN     236      // Ordenada na origem
#define POS(volt) (ORIGIN - (signed int) ((word) (volt)*SLOPE/QUOCIENT));

// Funo bloqueante de processamento das medidas
//
void transferServoSensors(word vpotMin, byte vpotImp, byte srv)
{
	signed int ang, ang_old;
    struct_velocity *ptr=velocityServo+srv;

    // Clculo da posio em graus
    ang=POS(vpotMin);                   // Clculo do ngulo associado  tenso medida
    if (ang<-127)     ang=-127;         // Truncagem do resultado de modo a caber numa varivel de 8 bits
    else if (ang>127) ang= 127;

    // Leitura da medio de posio mais antiga
    ang_old=ptr->positions[ptr->oldPosition];
    // A nova posio medida  armazenada no elemento mais antigo
    ptr->positions[ptr->oldPosition]=ang;
    // Actualizao do elemento do array de posies com a medio mais antiga
    ptr->oldPosition=(ptr->oldPosition+1)%N_POS;

    //ang=vpotMin;     // Clculo do ngulo associado  tenso medida

    sensors.servo[srv].position=(signed char) ang;      // Posio angular
    sensors.servo[srv].velocity=ang-ang_old;            // Velocidade (graus/N_POS ticks)
    sensors.servo[srv].current =vpotImp;                // Corrente
}

//#define KP  0.10  //0.17        // 0.10 ou 0.07 Para a gama -90 a 0 estes // 0.06 ou 0.06 -45=>+45 //  0.05    Destino: Binrio0
//#define KD  0.45  //0.90        // 0.15    0.00 valores so os ideais     // 0.06    0.00          //  0.00

void controller(void)
{
    bool controlOn;
    byte srv;
    signed char posReference;
    signed int error, error_dif;

    // Aplicao do controlador
    for (srv=0; srv<N_SERVOS; ++srv)  {

        // Determinao da posio desejada (referncia)
        posReference=pwmServo[srv].applyPosition;

        // Erro actual
   	    error=(signed int) posReference - sensors.servo[srv].position;
        if (error>-3 && error<3)  error=0;
   	    // Erro diferencial
        error_dif=error-pwmServo[srv].control.error_prev;
   	    // Actualizao do erro anterior
        pwmServo[srv].control.error_prev=error;

        // No caso do controlador ligado, aplicar compensao
        if ((sensors.sysStatus.controlOn[srv] = actuators.control[srv].controlOn
            && (actuators.control[srv].kp!=0 || actuators.control[srv].kd!=0 || actuators.control[srv].ki!=0)))  {

            // Clculo da nova posio a aplicar
            pwmServo[srv].control.angle += error*(signed int) actuators.control[srv].kp;      // Compensao proporcional
            pwmServo[srv].control.angle += error_dif*(signed int) actuators.control[srv].kd;  // Compensao derivativa
        
       	    // Limitao da posio a solicitar
            if (pwmServo[srv].control.angle<-(signed int) SERVO_RANGE*ANGLE_FACTOR)
                pwmServo[srv].control.angle=-(signed int) SERVO_RANGE*ANGLE_FACTOR;
       	    else if (pwmServo[srv].control.angle>(signed int) SERVO_RANGE*ANGLE_FACTOR)
                pwmServo[srv].control.angle= (signed int) SERVO_RANGE*ANGLE_FACTOR;
        }
        // Caso o controlador esteja desligado,
        // aplicar imediatamente a posio desejada
        else
            pwmServo[srv].control.angle=(signed int) posReference*ANGLE_FACTOR;


        // Converso de posio para PWM
        pwmServo[srv].pwm=PWM((signed char) (pwmServo[srv].control.angle/(signed int) 100));
    }
}
