#include "servo.h"
#include "can.h"

// Variveis da estrutura de informao temporal
bool timeTick=false;

// Ponto zero dos servos
signed int origin[N_SERVOS]=ORIGIN_STD;

// Macro conversora de ngulo para unidades de PWM
#define PWM(ang)    (byte) ((SERVO_RANGE-(signed int) (ang))*SERVO_RES/(2*(signed int) SERVO_RANGE))

// Estruturas externas
struct_calib calibServo[N_SERVOS];                // Dados de calibrao
volatile struct_pwmServo pwmServo[N_SERVOS];      // Dados dos actuadores
volatile struct_velocity velocityServo[N_SERVOS]; // Dados dos sensores




// Reinicializao da estrutura utilizada para calibrao
typedef enum {START_CALIB=false, NO_CALIB=true}  calib_type;

static void resetCalib(calib_type calib)
{
    byte srv;

    for (srv=0; srv<N_SERVOS; ++srv)  {
        // Reinicializao das variveis de estado
        calibServo[srv].n_measure  =0;
        calibServo[srv].voltage_sum=0;
        calibServo[srv].voltage_med=0;
        calibServo[srv].voltage_now=0;

        // Iniciao da calibrao
        // Se calib==false, uma calibrao ser feita no ciclo seguinte
        calibServo[srv].med_calculated=calib;
    }
}

// Inicializao das estruturas de dados
//
void initServoData(void)
{
    byte srv, i;

    // Reinicializao do estado de calibrao do servo
    resetCalib(NO_CALIB);

    for (srv=0; srv<N_SERVOS; ++srv)  {

        // Inicialmente todos os pwm esto desligados
        pwmServo[srv].pwmOn=false;

        // Inicializao das posies dos servos nos valores standard
        pwmServo[srv].applyPosition=homePosition[srv];
        pwmServo[srv].pwm=PWM(homePosition[srv]);

        // Inicializao dos arrays de posio
        for (i=0; i<N_POS; ++i)  {
            velocityServo[srv].positions[i]=homePosition[srv];
            velocityServo[srv].oldPosition=0;
        }
    }
}

static void applyMasterRequest(void)
{
    byte srv;
    signed char pos;

    // Esperando por uma ordem do Master
    while (!requestReceived);

    // Activao do PWM
    for (srv=0; srv<N_SERVOS; ++srv)  {

        // Definio da posio inicial
        pos=servoPos(actuators.servo[srv].position,srv);

        // Definio do PWM associado
        pwmServo[srv].pwm=PWM(pos);

        // Activao do sinal de PWM
        pwmServo[srv].pwmOn=true;
    }
}

// Execuo da calibrao dos servos
//
void calibration(void)
{
    byte srv;

    // Aplicao da posio pedida pelo Master
    applyMasterRequest();

    // Esperando 2 segundos para que a posio desejada seja atingida
    wait(2000);

    // Reset do estado de calibrao dos servos e iniciao da calibrao
    resetCalib(START_CALIB);

    for (srv=0; srv<N_SERVOS; ++srv)  {

        // Esperando o trmino do processo de medio
        while (!calibServo[srv].med_calculated);

        // Definio do ponto origem
        origin[srv]=ORIGIN(calibServo[srv].voltage_med,srv);
    }

    // Calibrao efectuada!
}

// ==========================================================================================
// Funes temporais
// ==========================================================================================

// Funcao para gerar um atraso (ms)
//
void wait(unsigned int value)
{
	unsigned int i, end=value/TICK_TIME;

	for (i=0; i<end; ++i)  {
		while (timeTick==0);
   		timeTick=0;
	}
}

// Aplicao de uma nova velocidade a um servo
//
signed char servoPos(signed char position, byte srv)
{
    if (position<-SERVO_RANGE)       position=-SERVO_RANGE;
    else if (position>SERVO_RANGE)   position= SERVO_RANGE;

    pwmServo[srv].applyPosition=position;

    return position;
}

void activatePWM(void)
{
    byte srv;

    for (srv=0; srv<N_SERVOS; ++srv)
        pwmServo[srv].pwmOn=true;
}

// Inicializao do Controlador
//
void initController(void)
{
    byte srv;

    for (srv=0; srv<N_SERVOS; ++srv)  {

        pwmServo[srv].control.angle=(double) actuators.servo[srv].position;
        pwmServo[srv].control.deltaD_prev=0.0;
        pwmServo[srv].control.error_prev=0;
        pwmServo[srv].control.y_prev=0;
        pwmServo[srv].control.y_prev2=0;
    }
}

//#define KP  0.10  //0.17        // 0.10 ou 0.07 Para a gama -90 a 0 estes // 0.06 ou 0.06 -45=>+45 //  0.05    Destino: Binrio0
//#define KD  0.45  //0.90        // 0.15    0.00 valores so os ideais     // 0.06    0.00          //  0.00

#define FILTER_N  1.0             // Nvel de filtragem da componente derivativa

void controller(void)
{
    byte srv;
    signed char posReference, y_now;
    signed int error, error_dif, y_vardif;
    double deltaP, deltaI, deltaD, kp, ki, kd;

    // Aplicao do controlador
    for (srv=0; srv<N_SERVOS; ++srv)  {

        posReference=pwmServo[srv].applyPosition;                       // Posio desejada (referncia)
        y_now = (signed int) sensors.servo[srv].position;   // Posio efectiva do servo

        // Erro actual
   	    error=(signed int) posReference - y_now;
        // Introduo da banda morta
        if (error>-2 && error<2)    error= 0;
        else if (error== 3)         error= 2;
        else if (error==-3)         error=-2;

   	    // Diferena do erro para a aco proporcional e variao da diferena da sada para a aco derivativa
        error_dif= error-pwmServo[srv].control.error_prev;

        y_vardif =(signed int) y_now-2*(signed int) pwmServo[srv].control.y_prev+(signed int) pwmServo[srv].control.y_prev2;

   	    // Actualizao das sadas/erros anteriores
        pwmServo[srv].control.error_prev=error;
        pwmServo[srv].control.y_prev2=pwmServo[srv].control.y_prev;
        pwmServo[srv].control.y_prev =y_now;

        // No caso do controlador ligado, aplicar compensao
        if ((sensors.sysStatus.controlOn[srv] = actuators.control[srv].controlOn
            && (actuators.control[srv].kp!=0 || actuators.control[srv].kd!=0 || actuators.control[srv].ki!=0)))  {

            ki = (double) actuators.control[srv].ki/(double) ANGLE_FACTOR;
            kp = (double) actuators.control[srv].kp/(double) ANGLE_FACTOR;
            kd = (double) actuators.control[srv].kd/(double) ANGLE_FACTOR;

            // Componentes do compensador
            deltaI = (double) error*ki;                     // Compensao integrativa
            deltaP = (double) error_dif*kp;                 // Compensao proporcional
            deltaD = (double) y_vardif*kd;                  // Compensao derivativa

            // Filtragem
//            deltaD = (pwmServo[srv].control.deltaD_prev*kd/FILTER_N + deltaD)/(1+kd/FILTER_N);
            pwmServo[srv].control.deltaD_prev=deltaD;       // Update da ltima variao de derivada

            // Clculo da nova posio a aplicar
            pwmServo[srv].control.angle += deltaI + deltaP + deltaD;
        }
        // Caso o controlador esteja desligado,
        // aplicar imediatamente a posio desejada
        else
            pwmServo[srv].control.angle=(double) posReference;

   	    // Limitao da posio a solicitar
        if ((signed int) pwmServo[srv].control.angle<-SERVO_RANGE)
            pwmServo[srv].control.angle=-(double) SERVO_RANGE;
   	    else if ((signed int) pwmServo[srv].control.angle>SERVO_RANGE)
            pwmServo[srv].control.angle= (double) SERVO_RANGE;

        // Converso de posio para PWM
        pwmServo[srv].pwm=PWM((signed char) pwmServo[srv].control.angle);

        // Activao da gerao de PWM
        // pwmServo[srv].pwmOn=true;
    }
}
