//*******************************************************************************************
// SERVO.C
//
// Source file com a definio dos device drivers para o controlo de um servomotor.
//
// (c) Milton Ruas <a21824@alunos.det.ua.pt>
//     DETUA & DEMUA, Nov/2005
//*******************************************************************************************

#include "pic2.h"
#include "can.h"

// ****** CONSTANTES ************************************************************************

#define VPOT_MEASURE    true
#define VPOT_3MOTORS    true

// O Timer 0  usado para fazer aquisio do sinal proveniente do potencimetro
#define T0_PRESC	16			// Prescaler (alterar o programa)
#define T0_OUT		120  	    // Perodo do Timer (us)
static const unsigned char kt0 = (unsigned char) (255-(unsigned int) F_CPU*T0_OUT/T0_PRESC+11);

// O Timer 1  utilizado para definir o perodo do PWM para o servo
#define T1_PRES 4				// Prescaler seleccionado na programacao do timer 1 (alterar funo initPit(...))
#define T1_OUT	20				// Perodo do Timer 1 (ms)
#define KT1		(65535-((unsigned int) F_CPU*1000/T1_PRES)*T1_OUT+1) // Constante de diviso p/ o Timer 1
static const unsigned char kt1H = (unsigned char) (KT1 >> 8);
static const unsigned char kt1L = (unsigned char) (KT1 & 0xFF);

static const unsigned int  kt1sec  = 65535-(unsigned int) F_CPU*T0_OUT/T1_PRES+1;           // Margem de segurana antes do rise up do PWM
static const unsigned int  kt1max  = KT1+(unsigned int) F_CPU*(DUTY_MIN-T0_OUT)/T1_PRES+1;  // Margem de segurana mnimo antes do fall down do PWM
static const unsigned int  kt1min  = KT1+(unsigned int) F_CPU*DUTY_MAX/T1_PRES+1;           // Margem de segurana mximo antes do fall down do PWM

// Parmetros do Timer2
#define T2_PRES		4       // Prescaler do Timer 2 (alterar funo initPit(..))
#define T2_POST     10      // Postscaler do Timer 2 (alterar funo initPit(..))
#define T2CON_STD   0b01001101;     // Postacaler 1:10, Prescaler 1:4, TMR2ON=true
#define T2CON_HF    0b00000100;     // Configurao para as interrupes de alta frequncia
#define T2_RES_FACTOR 10    // Factor de preciso do passo de variao do PWM
#define T2_OUT_HF   (unsigned char) (((unsigned int) DUTY_MAX-DUTY_MIN)*T2_RES_FACTOR/(unsigned int) SERVO_RES)
static const unsigned char kt2_1st = (unsigned char) ((unsigned int) DUTY_MIN*F_CPU/(T2_PRES*T2_POST)-4);
static const unsigned char kt2_hf  = (unsigned char) ((unsigned int) T2_OUT_HF*F_CPU/T2_RES_FACTOR-1);
// Nota importante para a verso 3.02 do compilador: o ajuste feito nas constantes servem para compensar o atraso provocado
// pela salvaguarda do contexto na chamada  int.

// ****** VARIVEIS ESTTICAS E GLOBAIS *****************************************************

static struct  {
    unsigned char which;
    struct {
	    unsigned char vpotImp;
	    unsigned int  vpotMin, vpotMin_prev;
    } measure[N_SERVOS];
} servo;

// Variveis da estrutura de informao temporal
bool timeTick=false;

// ****** DEFINIO DAS FUNES *************************************************************
void highISR(void);						// Prottipo da funo de servio a interrupes
void lowISR(void);

#pragma code InterruptVectorHigh = 0x08	//Quando ocorre um interrupt o programa "salta" para este endereo
void InterruptVectorHigh (void)
{
    _asm  goto highISR  _endasm
}

#pragma code InterruptVectorLow = 0x18	//Quando ocorre um interrupt o programa "salta" para este endereo
void InterruptVectorLow (void)
{
    _asm  goto lowISR  _endasm
}

#pragma code

static void initVars(void);

// ==========================================================================================
// Device Drivers
// ==========================================================================================

// Inicializacoes relativas ao microcontrolador
//
void initPic(void)
{
    byte i;

    // Inicializao das variveis internas do servo
    initVars();

    // Programao dos portos de entrada/sada
	TRISA=0b00000001;		// RA0 input / RA1..RA7 outputs
	TRISB=0b00000000;		// RB0...RB7 outputs
	TRISC=0b00001111;		// RC0...RC4 outputs; RC3..RC0 inputs

    // Inicializao dos pinos de sada
    PORTA=MUX_INPUT(servo.which)<<1;  // Seleco da entrada do multiplexer
    SERVO1_PIN=0;           // Pinos de PWM
    SERVO2_PIN=0;
    SERVO3_PIN=0;
    TESTPIN=0;              // Pinos de teste
    TESTPIN2=0;
    TESTPIN3=0;

	// Programao do Timer 2 (desligado e interrupes desactivadas)

    T2CON = T2CON_STD;      // Postacaler 1:10, Prescaler 1:4, TMR2ON=true
   	PR2   = kt2_1st;        // Perodo standard
	IPR1bits.TMR2IP=1;		// High priority interrupt
   	PIE1bits.TMR2IE=0;      // Interrupes do Timer2 desactivadas

    // Programacao do Timer 1

	T1CONbits.RD16=0;		// read/write in two 8-bit operations

	T1CONbits.T1CKPS1 = 1;
	T1CONbits.T1CKPS0 = 0;	// prescaler 1:4

    T1CONbits.T1OSCEN=0;    // Oscillator is shut-off
    T1CONbits.TMR1CS=0;     // Internal clock (Fosc/4)

    TMR1H = kt1H;  		    // Period
	TMR1L = kt1L;

	T1CONbits.TMR1ON=1;		// Timer 1 ON
	IPR1bits.TMR1IP=1;		// Low priority interrupt
	PIE1bits.TMR1IE=1;		// Timer1 interrupt enable

    // Programacao do Timer 0

	T0CONbits.T08BIT=1;		// 8 bit timer
    T0CONbits.T0CS=0;       // Internal instruction cycle clock

    T0CONbits.PSA=0;        // Prescaler is assigned
    T0CONbits.T0PS2=0;      // Prescaler => 1:16
    T0CONbits.T0PS1=1;
    T0CONbits.T0PS0=1;

    TMR0L=kt0;              // Contador inicializado
	T0CONbits.TMR0ON=1;		// Timer0 is On
	INTCON2bits.TMR0IP=1;	// High priority interrupt
    INTCONbits.TMR0IE=VPOT_MEASURE;     // Interrupes activadas

    // Programacao da ADC

	ADCON0=0b10000001;      // Fosc/64, channel 0(RA0), enable ADC converter
	ADCON1=0b01001110;      // Left justified; RA0 analog; the rest are digital

	IPR1bits.ADIP=1;		// Low Priority interrupt
	PIE1bits.ADIE=0;		// Low Priority interrupt

	// Inicializao das comunicaes CAN
	initCan();
    // Interrupes (recepo) de baixa prioridade
    IPR3bits.RXB0IP=0;
    IPR3bits.RXB1IP=0;
    // Activa os interrupts dos buffers recepcao CAN
    PIE3bits.RXB0IE=1;                                  // Interrupts Enable
    PIE3bits.RXB1IE=1;

    // Activacao global das interrupcoes
	RCONbits.IPEN=1;		// Interrupt priority enable
	INTCONbits.GIEH=1;		// Enable high priority interrupts
	INTCONbits.GIEL=1;      // Enable low priority interrupts

    // Esperando pela estabilizao dos sinais dos sensores
    for (i=0; i<2; ++i)  {
        timeTick=false;
        while (timeTick==false);
    }
}

// Inicializao de algumas variveis globais
//
static void initVars(void)
{
    unsigned char srv;

    servo.which=0;

    for (srv=0; srv<N_SERVOS; ++srv)  {

        servo.measure[srv].vpotMin=0xFFFF;
        servo.measure[srv].vpotMin_prev=0;
        servo.measure[srv].vpotImp=0;
    }

    // Inicializao das estruturas externas
    initServoData();
}

// ==========================================================================================
// Funes temporais
// ==========================================================================================

// Funcao para gerar um atraso (ms)
//
void wait(unsigned int value)
{
	unsigned int i, end=value/T1_OUT;

	for (i=0; i<end; ++i)  {
		while (timeTick==0);
   		timeTick=0;
	}
}

// ==========================================================================================
// Funes relativas a interrupes
// ==========================================================================================


// Actualizao das medidas
void updateServoMeasures(void);
void finalizeServoMeasures(void);
/*
// Transao CAN
typedef enum {
    NORMAL_TRANSACTION = 0,
    RESET_TRANSACTIONS = 1
} enum_transactionType;
static void canTransaction(enum_transactionType operation);
*/
// Rotina de servio s interrupes de alta prioridade
//
// O pragma seguinte indica que a rotina definida atende interrupes de alta prioridade,
// com uma seco de dados prpria com o nome isr_tmp, e que no  necessrio guardar o
// contexto da seco de dados geral.
#pragma interrupt highISR nosave=section(".tmpdata")
void highISR(void)
{
	unsigned char counter;
    unsigned int tmr1_val;

    // Interrupo responsvel por arrancar a ADC
    if (INTCONbits.TMR0IE && INTCONbits.TMR0IF)  {
        // Arranque do timer0 para provocar um delay antes da prxima medio
        TMR0L=kt0;

        // Apenas prossegue se existe tempo para executar as operaes associadas a este interrupt
        tmr1_val =((word) TMR1H << 8) | TMR1L;
        if ((!PIE1bits.TMR1IE || !PIR1bits.TMR1IF) && (!PIE1bits.TMR2IE || !PIR1bits.TMR2IF)
            && (tmr1_val<=kt1sec) && (tmr1_val<=kt1max || tmr1_val>=kt1min))  {

            TESTPIN3=1;

            // Arranque da ADC para leitura do sensor do servo
            PIR1bits.ADIF=0;
            PIE1bits.ADIE=1;
            ADCON0bits.GO=1;
        }
        INTCONbits.TMR0IF=0;
    }

    // Interrupo responsvel por ler os sensores e efectuar uma transao pelo CAN
    if (PIE1bits.ADIE && PIR1bits.ADIF)  {

        // Actualizao das medidas
        updateServoMeasures();

        TESTPIN3=0;

        PIE1bits.ADIE=0;
        PIR1bits.ADIF=0;
    }

    // Interrupo responsvel por baixar o PWM
    if (PIE1bits.TMR2IE && PIR1bits.TMR2IF)  {

        // O Timer2  redefinido de modo a provocar interrupes de alta frequncia
        T2CON=T2CON_HF;
        PR2=kt2_hf;

        counter=0;
        do  {
            while (PIR1bits.TMR2IF==0);
            PIR1bits.TMR2IF=0;
            
            if (SERVO1_PIN && counter>=pwmServo[0].pwm)  SERVO1_PIN=0;    // Desactivar PWM?
            if (SERVO2_PIN && counter>=pwmServo[1].pwm)  SERVO2_PIN=0;
            if (SERVO3_PIN && counter>=pwmServo[2].pwm)  SERVO3_PIN=0;
        } while (counter++<SERVO_RES);

        TESTPIN2=1;

        // Reset das definies do Timer2
        T2CON = T2CON_STD;      // Postacaler 1:10, Prescaler 1:4, TMR2ON=true
   	    PR2   = kt2_1st;        // Perodo standard
        PIE1bits.TMR2IE=0;      // Interrupes do Timer2 desactivadas

        // Clculo dos valores sensoriais definitivos para um ciclo de PWM
        finalizeServoMeasures();

        // Actualizao dos time Ticks temporais (apenas se o PWM estiver a ser gerado)
    	timeTick=true;

        // Reinicio das medies dos servos
        INTCONbits.TMR0IF=1;

        TESTPIN2=0;

        PIR1bits.TMR2IF=0;
	}

    // Interrupo responsvel por elevar o PWM
    if (PIE1bits.TMR1IE && PIR1bits.TMR1IF)  {

        TMR1H=kt1H;             // Inicializao do contador do Timer 1
        TMR1L=kt1L;

        // Inicializao do Timer2 para gerar o primeiro interrupt de alta frequncia
        TMR2=0;
       	PIR1bits.TMR2IF=0;
       	PIE1bits.TMR2IE=1;		// Activao das interrupes do Timer2

        // Se a gerao de PWM estiver autorizada, faz-lo:
        if (pwmServo[0].pwmOn)  SERVO1_PIN=1;           // Output for Servos activated
        if (pwmServo[1].pwmOn)  SERVO2_PIN=1;
        if (pwmServo[2].pwmOn)  SERVO3_PIN=1;

        INTCONbits.TMR0IF=1;    // Executar imediatamente uma medio do potencimetro

		PIR1bits.TMR1IF=0;      // Reset Timer1 interrupt flag
	}
}

// O cdigo definido no interior deste pragma indica que a seco de variveis temporrias
// a utilizar NO  a geral, mas sim a mesma que  utilizada pela RSI de alta prioridade.
#pragma tmpdata highISR_tmp
void updateServoMeasures(void)
{
    // Leitura da tenso presente no potencimetro do servo
    unsigned int vpot=((unsigned int) ADRESH<<2) | (ADRESL>>6);

	if (vpot<servo.measure[servo.which].vpotMin)
        servo.measure[servo.which].vpotMin=vpot;  // Actualizao do valor mnimo

    // Medio da largura do impulso de corrente
	if (vpot>=servo.measure[servo.which].vpotMin_prev+VPOT_IMPULSE_LIMIAR)
        servo.measure[servo.which].vpotImp++;

    // Transferncia da entrada do multiplexer para o servo seguinte
    if (VPOT_3MOTORS)  {
        if (++servo.which>=N_SERVOS)  servo.which=0;
        PORTA=MUX_INPUT(servo.which)<<1;
    }
}

// Clculo dos valores sensoriais definitivos para um ciclo de PWM
//
void finalizeServoMeasures(void)
{
    byte srv, i;
    word vpotMin, val;
    byte vpotImp;

    word current_sum;
	signed int ang, ang_old;
    struct_velocity *ptr;


  	// Passagem dos valores finais
    for (srv=0; srv<N_SERVOS; ++srv)  {

        val=servo.measure[srv].vpotMin;
        vpotMin=val*5-val*14/125;               // Converso para tenso (0..5000)
        vpotImp=servo.measure[srv].vpotImp;

	    // Reinicializao das variveis auxiliares
	    servo.measure[srv].vpotMin_prev=servo.measure[srv].vpotMin;
	    servo.measure[srv].vpotMin=0xFFFF;
	    servo.measure[srv].vpotImp=0;

        // Clculo da posio em graus
        ang=POS(vpotMin,srv);                   // Clculo do ngulo associado  tenso medida
        if (ang<-127)     ang=-127;         // Truncagem do resultado de modo a caber numa varivel de 8 bits
        else if (ang>127) ang= 127;

        ptr=velocityServo+srv;
        // Leitura da posio e da corrente mais antiga
        ang_old=ptr->positions[ptr->oldPosition];
        // Soma das correntes medidas nas ltimas N_POS medies
        for (i=0,current_sum=0; i<N_POS; ++i)  current_sum+=(word) ptr->current[i];

        // A nova corrente medida  armazenda no elemento mais antigo
        ptr->current[ptr->oldPosition]=vpotImp;
        // A nova posio medida  armazenada no elemento mais antigo
        ptr->positions[ptr->oldPosition]=ang;
        // Actualizao do elemento do array de posies com a medio mais antiga
        ptr->oldPosition=(ptr->oldPosition+1)%N_POS;

        // Actualizao das estruturas globais
        sensors.servo[srv].position=(signed char) ang;          // Posio angular
        sensors.servo[srv].velocity=ang-ang_old;                // Velocidade (graus/N_POS ticks)
        sensors.servo[srv].current =current_sum/N_POS;          // Corrente mdia deslizante
    }
}
#pragma tmpdata

// Rotina de atendimento s interrupes de baixa prioridade
#pragma interruptlow lowISR
void lowISR(void)
{
    TESTPIN4=1;
    if (PIR3bits.RXB0IF)  {

        receiveCanMessage(0);
        sendCanMessage(0);
    }

    if (PIR3bits.RXB1IF)  {

        receiveCanMessage(1);
        sendCanMessage(1);
    }
    TESTPIN4=0;
}

/*
static void canTransaction(enum_transactionType operation)
{
    // Mquina de estados na troca de mensagens CAN
    #define N_OPERATIONS    4
    static enum {
        RECV_MESSAGE1 = 0,
        RECV_MESSAGE2 = 1,
        SEND_MESSAGE1 = 2,
        SEND_MESSAGE2 = 3
    } can_op=RECV_MESSAGE1;
    bool res=true;

    // Caso tenha sido solicitado um reset das comunicaes,
    // a mquina de estados interrompe a sequncia e volta ao incio.
    if (operation==RESET_TRANSACTIONS)  {
        can_op=RECV_MESSAGE1;
        return;
    }

    switch (can_op)  {
        // Recepo das mensagens de actualizao dos actuadores
        case RECV_MESSAGE1: res=receiveCanMessage(0);
                            break;
        case RECV_MESSAGE2: res=receiveCanMessage(1);
                            break;
        // Envio da primeira mensagem de actualizao dos actuadores
        case SEND_MESSAGE1: sendCanMessage(0);
                            break;
        // Envio da segunda mensagem de actualizao dos actuadores
        case SEND_MESSAGE2: sendCanMessage(1);
                            break;
    }

    // Apenas passa para a operao seguinte, se a actual (res) teve sucesso;
    // e apenas muda o slave a operar, quando a mquina de estados volta ao incio.
    if (res)  can_op=(can_op+1)%N_OPERATIONS;
}
*/
