#include "servo.h"
#include "pic2.h"
#include "can.h"

// Estrutura com os dados do controlador
typedef struct {
    double angle;               // ngulo a solicitar
    double deltaD_prev;         // Variao derivativa do ciclo anterior
    signed int  error_prev;     // Erro do ciclo anterior
    signed char y_prev;         // Sinal de feedback (potencimetro) do ciclo anterior
    signed char y_prev2;        // Sinal de feedback de h 2 ciclos atrs
} struct_controller;

// Estrutura com os dados do controlador
typedef struct  {
    signed char finalPosition;      // Posio final a atingir
    struct_controller controller;   // Dados do controlador
} struct_servoControl;
static struct_servoControl servoControl[N_SERVOS];      // Dados dos actuadores


// Execuo da calibrao dos servos
//
void calibration(void)
{
    byte i, count;
    signed int med;
    unsigned int reaction_sum[N_FORCE_SENSORS];
    signed int position_sum[N_SERVOS];
    signed char posReq[N_SERVOS];

    // Esperando por uma ordem do Master
    while (!canRequest);

    // Activao do PWM
    for (i=0; i<N_SERVOS; ++i)  {

        // Posio solicitada
        posReq[i]=actuators.servo[i].position;

        // Definio da posio inicial
        servoActuation(posReq[i],i);

        // Activao do sinal de PWM
        statusPWM(true,i);

        // Inicializao do array
        position_sum[i]=0;
    }
    // Inicializao das variveis locais
    for (i=0; i<N_FORCE_SENSORS; ++i)  reaction_sum[i]=0;

    // Esperando 2 segundos para que a posio desejada seja atingida
    wait(2000);

    // Execuo de 25 medidas
    for (count=0; count<25; ++count)  {
        // Esperando pela passagem de um tick;
        waitTick();

        // Acumulao das posies medidas
        for (i=0; i<N_SERVOS; ++i)
            position_sum[i] += servo_interface[i].position;
        // Acumulao da sada dos sensores de fora
        for (i=0; i<N_FORCE_SENSORS; ++i)
            reaction_sum[i] += sensors.reaction[i];
    }

    // Correco dos desvios da medio do potencimetro
    for (i=0; i<N_SERVOS; ++i)  {
        med = (signed int) (position_sum[i]/25);
        origin[i] += (signed char) ((signed int) posReq[i]-med);

        // Activao dos filtros
        statusFilter(true,i);
    }
    // Correco dos desvios dos sensores de fora
    for (i=0; i<N_FORCE_SENSORS; ++i)  {
        med = (signed int) (reaction_sum[i]/25);

        // Offset determinado
        reaction0[i] += (unsigned char) med - 10;
    }
    // Calibrao efectuada!
}



// Movimentao do servo para a posio solicitada
//
signed char servoRequest(signed char position, byte srv)
{
    // Correco da posio solicitada
    position=limitPosition(position);

    // Posio a solicitar
    servoControl[srv%N_SERVOS].finalPosition=position;

    return position;
}

// ==========================================================================================
// Funes temporais
// ==========================================================================================

// Inicializao do Controlador
//
void initController(void)
{
    byte srv;
    struct_controller *cont;

    for (srv=0; srv<N_SERVOS; ++srv)  {

        cont = &(servoControl[srv].controller);

        servoControl[srv].finalPosition=actuators.servo[srv].position;
        cont->angle=(double) servoControl[srv].finalPosition;
        cont->deltaD_prev=0.0;
        cont->error_prev=0;
        cont->y_prev=0;
        cont->y_prev2=0;
    }
}

//#define KP  0.10  //0.17        // 0.10 ou 0.07 Para a gama -90 a 0 estes // 0.06 ou 0.06 -45=>+45 //  0.05    Destino: Binrio0
//#define KD  0.45  //0.90        // 0.15    0.00 valores so os ideais     // 0.06    0.00          //  0.00

#define FILTER_N  1.0             // Nvel de filtragem da componente derivativa

void controller(void)
{
    byte srv;
    signed char posReference, y_now;
    signed int error, error_dif, y_vardif;
    double deltaP, deltaI, deltaD, kp, ki, kd;
    struct_controller *cont;

    // Aplicao do controlador
    for (srv=0; srv<N_SERVOS; ++srv)  {

        // Ponteiro para o servo em causa
        cont = &(servoControl[srv].controller);

        posReference=servoControl[srv].finalPosition;       // Posio desejada (referncia)
        y_now = servo_interface[srv].position;              // Posio efectiva do servo

        // Erro actual
   	    error=(signed int) posReference - (signed int) y_now;
        // banda morta
        if (error>=-2 && error<=2)  error= 0;
        else if (error== 3)         error= 2;
        else if (error==-3)         error=-2;

   	    // Diferena do erro para a aco proporcional e variao da diferena da sada para a aco derivativa
        error_dif= error-cont->error_prev;
        y_vardif =(signed int) y_now-2*(signed int) cont->y_prev+(signed int) cont->y_prev2;

   	    // Actualizao das sadas/erros anteriores
        cont->y_prev2=cont->y_prev;
        cont->y_prev =y_now;
        cont->error_prev=error;

        // Controlador ligado?
        sensors.sysStatus.controlOn[srv] = actuators.control[srv].controlOn
            && (actuators.control[srv].kp!=0 || actuators.control[srv].kd!=0 || actuators.control[srv].ki!=0);
        // No caso do controlador ligado, aplicar compensao
        if (sensors.sysStatus.controlOn[srv])  {

            ki = (double) actuators.control[srv].ki/(double) K_FACTOR;
            kp = (double) actuators.control[srv].kp/(double) K_FACTOR;
            kd = (double) actuators.control[srv].kd/(double) K_FACTOR;

            // Componentes do compensador
            deltaI = (double) error*ki;                     // Compensao integrativa
            deltaP = (double) error_dif*kp;                 // Compensao proporcional
            deltaD = (double) y_vardif*kd;                  // Compensao derivativa

            // Filtragem
//            deltaD = (pwmServo[srv].control.deltaD_prev*kd/FILTER_N + deltaD)/(1+kd/FILTER_N);
            cont->deltaD_prev=deltaD;   // Update da ltima variao de derivada

            // Clculo da nova posio a aplicar
            cont->angle += deltaI + deltaP + deltaD;
        }
        // Caso o controlador esteja desligado,
        // aplicar imediatamente a posio desejada
        else
            cont->angle=(double) posReference;

   	    // Limitao da posio a solicitar
        if ((signed int) cont->angle<-SERVO_RANGE)      cont->angle=-(double) SERVO_RANGE;
   	    else if ((signed int) cont->angle>SERVO_RANGE)  cont->angle= (double) SERVO_RANGE;

        // Aplicao do movimento
        servoActuation((signed char) cont->angle,srv);
    }
}
